/**
 * ==============================================
 *  LibReDE : Library for Resource Demand Estimation
 * ==============================================
 *
 * (c) Copyright 2013-2014, by Simon Spinner and Contributors.
 *
 * Project Info:   http://www.descartes-research.net/
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.prisma.agent.wildfly.extension;

import java.util.List;

import org.jboss.as.controller.AbstractAddStepHandler;
import org.jboss.as.controller.OperationContext;
import org.jboss.as.controller.OperationFailedException;
import org.jboss.as.controller.PathAddress;
import org.jboss.as.controller.ServiceVerificationHandler;
import org.jboss.as.controller.registry.Resource;
import org.jboss.as.server.AbstractDeploymentChainStep;
import org.jboss.as.server.DeploymentProcessorTarget;
import org.jboss.dmr.ModelNode;
import org.jboss.dmr.Property;
import org.jboss.msc.service.ServiceBuilder;
import org.jboss.msc.service.ServiceController;
import org.jboss.msc.service.ServiceController.Mode;
import org.jboss.msc.value.InjectedValue;

import tools.descartes.prisma.agent.wildfly.deployment.AddDependenciesDeploymentProcessor;
import tools.descartes.prisma.agent.wildfly.deployment.ClientInstrumentationDeploymentProcessor;
import tools.descartes.prisma.agent.wildfly.deployment.ServerInstrumentationDeploymentProcessor;
import tools.descartes.prisma.agent.wildfly.service.FilterService;
import tools.descartes.prisma.agent.wildfly.service.InstrumentationService;
import tools.descartes.prisma.agent.wildfly.service.ModelExtractionService;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;
import tools.descartes.prisma.agent.wildfly.service.ParameterCharacterizationService;

/**
 * Handler responsible for adding the subsystem resource to the model
 *
 * @author <a href="kabir.khan@jboss.com">Kabir Khan</a>
 */
class SubsystemAdd extends AbstractAddStepHandler {

	static final SubsystemAdd INSTANCE = new SubsystemAdd();

	private SubsystemAdd() {
		super(SubsystemDefinition.REPOSITORY_HOST, SubsystemDefinition.REPOSITORY_USER,
				SubsystemDefinition.REPOSITORY_PASSWORD, SubsystemDefinition.SCOPE);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void populateModel(ModelNode operation, ModelNode model) throws OperationFailedException {
		SubsystemDefinition.REPOSITORY_HOST.validateAndSet(operation, model);
		SubsystemDefinition.REPOSITORY_USER.validateAndSet(operation, model);
		SubsystemDefinition.REPOSITORY_PASSWORD.validateAndSet(operation, model);
		SubsystemDefinition.SCOPE.validateAndSet(operation, model);
	}

	@Override
	protected void performRuntime(OperationContext context, ModelNode operation, ModelNode model,
			ServiceVerificationHandler verificationHandler, List<ServiceController<?>> newControllers)
			throws OperationFailedException {
		if (context.isBooting()) {
			System.err.println("Add prisma deployment processors");
			context.addStep(new AbstractDeploymentChainStep() {
				public void execute(DeploymentProcessorTarget processorTarget) {
					processorTarget.addDeploymentProcessor(SubsystemExtension.SUBSYSTEM_NAME,
							AddDependenciesDeploymentProcessor.PHASE, AddDependenciesDeploymentProcessor.PRIORITY,
							new AddDependenciesDeploymentProcessor());
					processorTarget.addDeploymentProcessor(SubsystemExtension.SUBSYSTEM_NAME,
							ServerInstrumentationDeploymentProcessor.PHASE,
							ServerInstrumentationDeploymentProcessor.PRIORITY,
							new ServerInstrumentationDeploymentProcessor());
					processorTarget.addDeploymentProcessor(SubsystemExtension.SUBSYSTEM_NAME,
							ClientInstrumentationDeploymentProcessor.PHASE,
							ClientInstrumentationDeploymentProcessor.PRIORITY,
							new ClientInstrumentationDeploymentProcessor());
				}
			}, OperationContext.Stage.RUNTIME);
		}
		System.err.println("Start services");
		final ModelNode fullModel = Resource.Tools.readModel(context.readResource(PathAddress.EMPTY_ADDRESS));
		String repositoryHost = fullModel.get(SubsystemExtension.REPOSITORY_HOST_ATTRIBUTE).asString();
		String repositoryUser = fullModel.get(SubsystemExtension.REPOSITORY_USER_ATTRIBUTE).asString();
		String repositoryPassword = fullModel.get(SubsystemExtension.REPOSITORY_PASSWORD_ATTRIBUTE).asString();
		String scopePath = fullModel.get(SubsystemExtension.SCOPE_ATTRIBUTE).asString();
		
		ModelExtractionService.INSTANCE.init(repositoryHost, repositoryUser, repositoryPassword, scopePath);
		MonitoringService.INSTANCE.init(repositoryHost, repositoryUser, repositoryPassword, scopePath);

		ServiceBuilder<MonitoringService> monitoringControllerBuilder = context.getServiceTarget()
				.addService(MonitoringService.SERVICE_NAME, MonitoringService.INSTANCE).setInitialMode(Mode.ACTIVE);
		if (fullModel.hasDefined(SubsystemExtension.PARAM_CHARACTERIZATION)) {
			for (Property prop : fullModel.get(SubsystemExtension.PARAM_CHARACTERIZATION).asPropertyList()) {
				String name = prop.getName();
				InjectedValue<ParameterCharacterizationService> v = new InjectedValue<>();
				monitoringControllerBuilder.addDependency(MonitoringService.PARAM_CHARACTERIZATION_NAME.append(name),
						ParameterCharacterizationService.class, v);
				MonitoringService.INSTANCE.getParameterCharacterizations().add(v);
			}
		}

		ServiceController<MonitoringService> monitoringController = monitoringControllerBuilder.install();

		ServiceBuilder<ModelExtractionService> extractionControllerBuilder = context.getServiceTarget()
				.addService(ModelExtractionService.SERVICE_NAME, ModelExtractionService.INSTANCE)
				.setInitialMode(Mode.ACTIVE).addDependency(MonitoringService.SERVICE_NAME, MonitoringService.class,
						ModelExtractionService.INSTANCE.getMonitoringService());
		if (fullModel.hasDefined(SubsystemExtension.FILTER)) {
			for (Property prop : fullModel.get(SubsystemExtension.FILTER).asPropertyList()) {
				String name = prop.getName();
				InjectedValue<FilterService> v = new InjectedValue<>();
				extractionControllerBuilder.addDependency(MonitoringService.FILTER_NAME.append(name),
						FilterService.class, v);
				ModelExtractionService.INSTANCE.getFilters().add(v);
			}
		}

		ServiceController<ModelExtractionService> extractionController = extractionControllerBuilder.install();

		InstrumentationService instrumentationService = new InstrumentationService();
		ServiceController<InstrumentationService> instrumentationController = context.getServiceTarget()
				.addService(InstrumentationService.SERVICE_NAME, instrumentationService).setInitialMode(Mode.ACTIVE)
				.install();

		if (newControllers != null) {
			newControllers.add(instrumentationController);
			newControllers.add(monitoringController);
			newControllers.add(extractionController);
		}
	}
}
