/**
 * ==============================================
 *  LibReDE : Library for Resource Demand Estimation
 * ==============================================
 *
 * (c) Copyright 2013-2014, by Simon Spinner and Contributors.
 *
 * Project Info:   http://www.descartes-research.net/
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.prisma.agent.wildfly.service;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import tools.descartes.prisma.agent.wildfly.Constants;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.parameters.ValueCharacterization;
import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.core.impl.RabbitMQMessageBus;
import tools.descartes.prisma.core.scopes.ApplicationScope;
import tools.descartes.prisma.model.scope.ScopePackage;
import tools.descartes.prisma.model.sensor.ComponentSensor;


public class MonitoringService extends AgentController implements Service<MonitoringService> {
	
	private static final Logger log = Logger.getLogger(MonitoringService.class);
	
	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "monitoring");
	
	public static final ServiceName FILTER_NAME = SERVICE_NAME.append("filter");
	
	public static final ServiceName PARAM_CHARACTERIZATION_NAME = SERVICE_NAME.append("filter");
	
	public static final MonitoringService INSTANCE = new MonitoringService();
	
	private final ConcurrentHashMap<Long, Long> threads = new ConcurrentHashMap<Long, Long>();
	
	private List<InjectedValue<ParameterCharacterizationService>> paramCharacterizations = new CopyOnWriteArrayList<>();

	private Map<String, Map<String, ValueCharacterization>> valueCharacterizations = new HashMap<>();

	private final ConcurrentHashMap<String, OperationRecord> monitoredEndpoints = new ConcurrentHashMap<>();

	private final ConcurrentHashMap<String, ComponentRecord> components = new ConcurrentHashMap<>();

	private ApplicationScope applicationScope;

	private static final ThreadLocal<Deque<InvocationRecord>> callStack = new ThreadLocal<Deque<InvocationRecord>>() {
		protected Deque<InvocationRecord> initialValue() {
			return new ArrayDeque<InvocationRecord>(10);
		};
	};

	private class OutputWriter implements Runnable {

		private PrintWriter out;

		public OutputWriter() throws IOException {
			out = new PrintWriter(new FileWriter(new File("./operations.csv"), true));
		}

		public void run() {
			try {
				long currentTime = System.currentTimeMillis();
				collectMonitoringData(out, (currentTime / 1000.0));
				out.flush();
			} catch (Throwable ex) {
				log.error("Error collecting monitoring data.", ex);
			}
		};
	};

	public void init(String repositoryHost, String repositoryUser, String repositoryPassword, String scopePath) {
		MessageBus messageBus = new RabbitMQMessageBus(repositoryHost, repositoryUser, repositoryPassword);
		ModelRepository repository = new CDOModelRepository(repositoryHost, repositoryUser, repositoryPassword,
				messageBus);
		super.init(repository, messageBus, scopePath,
				getLocalHost().toString() + "_" + UUID.randomUUID().toString());
	}

	@Override
	public MonitoringService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext ctx) throws StartException {
		ScopePackage.eINSTANCE.eClass();

		for (InjectedValue<ParameterCharacterizationService> param : paramCharacterizations) {
			addValueCharacterization(param.getValue().createValueCharacterization());
		}
		log.info("MonitoringService started");

		try {
			start().get();
			applicationScope = (ApplicationScope) getContainerScope();
			scheduleAtIntervalBounds(new OutputWriter(), Constants.MONITORING_INTERVAL_IN_MILLIS,
					TimeUnit.MILLISECONDS);
		} catch (InterruptedException | ExecutionException | IOException e) {
			log.error("Error starting agent.", e);
		}
	}

	@Override
	public void stop(StopContext arg0) {
		try {
			stop().get();
		} catch (InterruptedException | ExecutionException e) {
			log.error("Error stopping agent.", e);
		}
	}
	
	private void collectMonitoringData(PrintWriter out, double timestampInSec) {
		for (OperationRecord curEndpoint : monitoredEndpoints.values()) {
			for (InvocationRecord curInvocation : curEndpoint.getInvocations()) {
				String entityName = curInvocation.toString();
				long inInvocationCount = curInvocation.getInvocationCount();
				if (inInvocationCount <= 0) {
					continue;
				}
				long inExecutionTime = curInvocation.getExecutionTime();
				long inResidenceTime = curInvocation.getResidenceTime();
				// Send to message bus
				sendData(curInvocation.getThroughputSensor(), entityName, timestampInSec,
						inInvocationCount);
				sendData(curInvocation.getResponseTimeSensor(), entityName, timestampInSec,
						inExecutionTime);
				sendData(curInvocation.getResidenceTimeSensor(), entityName, timestampInSec, inResidenceTime);

				// Write to log file
				out.printf("%f, %s, , %d, %d, %d, %s\n", timestampInSec, curInvocation.toString(), inInvocationCount,
						inExecutionTime, inResidenceTime, curInvocation.getAdditionalInformation());

				for (OperationRecord curOutgoingEndpoint : curInvocation.getCalledEndpoints()) {
					for (InvocationRecord curOutgoingInvocation : curOutgoingEndpoint.getInvocations()) {
						String externalCallEntityName = entityName + "; " + curOutgoingInvocation.getInvocationCount();
						long outInvocationCount = curOutgoingInvocation.getInvocationCount();
						long outExecutionTime = curOutgoingInvocation.getExecutionTime();

						// Send to message bus
						sendData(curOutgoingInvocation.getThroughputSensor(), externalCallEntityName, timestampInSec,
								outInvocationCount);
						sendData(curOutgoingInvocation.getResponseTimeSensor(), externalCallEntityName, timestampInSec,
								outExecutionTime);

						// Write to log file
						out.printf("%f, %s, %s, %d, %d, %s\n", timestampInSec, curInvocation.toString(),
								curOutgoingInvocation.toString(), outInvocationCount, outExecutionTime,
								curOutgoingInvocation.getAdditionalInformation());
					}
				}
			}
		}
	}

	private void sendData(ComponentSensor<?> sensor, String entity, double timestamp, double value) {
		if (sensor != null) {
			sendMonitoringData(applicationScope.getApplicationStatisticsOutputPort(), sensor, entity,
				new double[] { timestamp }, new double[] { value });
		}
	}
	
	public long getLastCpuTime(long threadId) {
		Long val = threads.get(threadId);
		return (val == null) ? 0 : val;
	}
	
	public void setLastCpuTime(long threadId, long cpuTimeInNanos) {
		threads.put(threadId, cpuTimeInNanos);
	}
	
	public List<InjectedValue<ParameterCharacterizationService>> getParameterCharacterizations() {
		return paramCharacterizations;
	}

	public ComponentRecord addAlias(String alias, ComponentRecord record) {
		return components.put(alias, record);
	}

	public void addValueCharacterization(ValueCharacterization charac) {
		System.err.println("Value characterization added");
		Map<String, ValueCharacterization> ops = valueCharacterizations.get(charac.getInterfaceName());
		if (ops == null) {
			ops = new HashMap<>();
			valueCharacterizations.put(charac.getInterfaceName(), ops);
		}
		ops.put(charac.getOperationName(), charac);
	}

	public OperationRecord createOperationRecord(ComponentRecord component, String interf, String protocol,
			String operation) {
		return createOperationRecord(null, component, interf, protocol, operation);
	}

	public OperationRecord createOperationRecord(InvocationRecord caller, ComponentRecord component, String interf,
			String protocol, String operation) {
		ValueCharacterization valueCharac = null;
		Map<String, ValueCharacterization> charac = valueCharacterizations.get(interf);
		if (charac != null) {
			valueCharac = charac.get(operation);
		}
		return new OperationRecord(this, caller, component, interf, protocol, operation, valueCharac);
	}

	public OperationRecord requireOperation(ComponentRecord component, String interf, String protocol,
			String operation) {
		String endpointIdentifier = OperationRecord.getEndpointIdentification(component, interf, protocol, operation);
		OperationRecord record = monitoredEndpoints.get(endpointIdentifier);
		if (record == null) {
			record = createOperationRecord(component, interf, protocol, operation);
			if (monitoredEndpoints.putIfAbsent(endpointIdentifier, record) != null) {
				// a different thread has changed it in the mean time
				record = monitoredEndpoints.get(endpointIdentifier);
			} else {
				ModelExtractionService.INSTANCE.newOperationRecord(record);
			}
		}
		return record;
	}

	public ComponentRecord getComponent(String identifier) {
		return components.get(identifier);
	}

	public ComponentRecord requireComponent(String serverNode, String application, String module, String component) {
		String componentIdentifier = ComponentRecord.getComponentIdentifier(serverNode, application, module, component);
		ComponentRecord record = components.get(componentIdentifier);
		if (record == null) {
			record = new ComponentRecord(serverNode, application, module, component);
			if (components.putIfAbsent(componentIdentifier, record) != null) {
				// a different thread has changed it in the mean time
				record = components.get(componentIdentifier);
			} else {
				ModelExtractionService.INSTANCE.newComponentRecord(record);
			}
		}
		return record;
	}

	public Collection<OperationRecord> getEndpoints() {
		return monitoredEndpoints.values();
	}

	public Deque<InvocationRecord> getCallStack() {
		return callStack.get();
	}

}
